#cython: infer_types=True

"""
A Python wrapper around CityHash, a fast non-cryptographic hashing algorithm
"""

__author__      = "Alexander [Amper] Marshalov"
__email__       = "alone.amper+cityhash@gmail.com"
__version__     = '0.1.0'
__all__         = ["CityHash64",
                   "CityHash64WithSeed",
                   "CityHash64WithSeeds",
                   "CityHash128",
                   "CityHash128WithSeed",
                  ]

cdef extern from * nogil:
    ctypedef unsigned long int uint32_t
    ctypedef unsigned long long int uint64_t

cdef extern from "<utility>" namespace "std":
    cdef cppclass pair[T, U]:
        T first
        U second
        pair()
        pair(pair&)
        pair(T&, U&)
        bint operator == (pair&, pair&)
        bint operator != (pair&, pair&)
        bint operator <  (pair&, pair&)
        bint operator >  (pair&, pair&)
        bint operator <= (pair&, pair&)
        bint operator >= (pair&, pair&)

cdef extern from "city.h" nogil:
    ctypedef uint32_t uint32
    ctypedef uint64_t uint64
    ctypedef pair uint128
    cdef uint64  c_Uint128Low64 "Uint128Low64" (uint128& x)
    cdef uint64  c_Uint128High64 "Uint128High64" (uint128& x)
    cdef uint64  c_CityHash64 "CityHash64" (char *buff, size_t len)
    cdef uint64  c_CityHash64WithSeed "CityHash64WithSeed" (char *buff, size_t len, uint64 seed)
    cdef uint64  c_CityHash64WithSeeds "CityHash64WithSeeds" (char *buff, size_t len, uint64 seed0, uint64 seed1)
    cdef uint128[uint64,uint64] c_CityHash128 "CityHash128" (char *s, size_t len)
    cdef uint128[uint64,uint64] c_CityHash128WithSeed "CityHash128WithSeed" (char *s, size_t len, uint128[uint64,uint64] seed)


cdef const char* _chars(basestring s):
    if isinstance(s, unicode):
        s = s.encode('utf8')
    return s


cpdef CityHash64(basestring buff):
    """
        Description: Hash function for a byte array.
    """
    cdef const char* array = _chars(buff)
    return c_CityHash64(array, len(array))

cpdef CityHash64WithSeed(basestring buff, uint64 seed=0L):
    """
        Description: Hash function for a byte array. For convenience, a 64-bit seed is also
                     hashed into the result.
    """
    cdef const char* array = _chars(buff)
    return c_CityHash64WithSeed(array, len(array), seed)

cpdef CityHash64WithSeeds(basestring buff, uint64 seed0=0L, uint64 seed1=0L):
    """
        Description: Hash function for a byte array.  For convenience, two seeds are also
                     hashed into the result.
    """
    cdef const char* array = _chars(buff)
    return c_CityHash64WithSeeds(array, len(array), seed0, seed1)

cpdef CityHash128(basestring buff):
    """
        Description: Hash function for a byte array.
    """
    cdef const char* array = _chars(buff)
    cdef pair[uint64,uint64] result = c_CityHash128(array, len(array))
    return 0x10000000000000000L * long(result.first) + long(result.second)


cpdef CityHash128WithSeed(basestring buff, seed=0L):
    """
        Description: Hash function for a byte array.  For convenience, a 128-bit seed is also
                     hashed into the result.
    """
    cdef uint64 seed_0 = seed >> 64
    cdef uint64 seed_1 = seed & ((1 << 64) - 1)
    cdef const char* array = _chars(buff)
    cdef pair[uint64,uint64] tseed
    tseed.first, tseed.second = seed_0, seed_1
    cdef pair[uint64,uint64] result = c_CityHash128WithSeed(array, len(array), tseed)
    return 0x10000000000000000L * long(result.first) + long(result.second)



__author__  = "Alexander [Amper] Marshalov"
__email__   = "alone.amper+cityhash@gmail.com"
__icq__     = "87-555-3"
__jabber__  = "alone.amper@gmail.com"
__twitter__ = "amper"
__url__     = "http://amper.github.com/cityhash"

from setuptools import setup, Extension
from setuptools.dist import Distribution
from pkg_resources import resource_string
from Cython.Distutils import build_ext


class BinaryDistribution(Distribution):
    """
    Subclass the setuptools Distribution to flip the purity flag to false.
    See http://lucumr.pocoo.org/2014/1/27/python-on-wheels/
    """
    def is_pure(self):
        # TODO: check if this is still necessary with Python v2.7
        return False


class build_ext_subclass(build_ext):
    """
    This class is an ugly hack to a problem that arises when one must force
    a compiler to use specific flags by adding to the environment somethiing
    like the following:

        CXX="clang --some_flagA --some_flagB -I/usr/bin/include/mylibC"

    (as opposed to setting CXXFLAGS). Distutils in that case will complain
    that it cannot run the entire command as given because it is not
    found as an executable (specific error message is: "unable to execute...
    ... no such file or directory").

    This subclass of ``build_ext`` will extract the compiler name from the
    command line and insert any remaining arguments right after it.
    """
    def build_extensions(self):
        ccm = self.compiler.compiler
        if ' ' in ccm[0]:
            self.compiler.compiler = ccm[0].split(' ') + ccm[1:]
        cxx = self.compiler.compiler_cxx
        if ' ' in cxx[0]:
            self.compiler.compiler_cxx = cxx[0].split(' ') + cxx[1:]
        build_ext.build_extensions(self)


CXXFLAGS = u"""
-O3
-msse4.2
-Wno-unused-value
-Wno-unused-function
""".split()

VERSION = '0.1.0'
URL = "https://github.com/escherba/python-cityhash"

setup(
    version=VERSION,
    description="Python-bindings for CityHash",
    author="Alexander [Amper] Marshalov",
    author_email="alone.amper+cityhash@gmail.com",
    url=URL,
    download_url=URL + "/tarball/master/" + VERSION,
    name='cityhash',
    license='MIT',
    cmdclass={'build_ext': build_ext_subclass},
    ext_modules=[Extension("cityhash", ["src/city.cc", "src/cityhash.pyx"],
                           language="c++",
                           extra_compile_args=CXXFLAGS,
                           include_dirs=['include'])],
    keywords=['hash', 'hashing', 'cityhash'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: C++',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities'
    ],
    long_description=resource_string(__name__, 'README.rst'),
    distclass=BinaryDistribution,
)
